#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
All you need to shoot yourself an arrow in the knee : preconfigured class
that seems harmless and seems to be powerfull. 

Short bows = one operator 
long bows = one quiver
Cross bow = algebrae + other quiver
Japanese bows : algebrae
European bows (common names) : proposed standard behaviour
I think I am already short of ideas

"""
__all__ = [ 'Hankyu', 'Daikyu', "vdict","sdict", 'mdict' ]

from .trait import (
        InclusiveAdder, InclusiveSubber, ExclusiveMuler, 
        Iterator, Searchable 
    )
from .quiver import LinearAlgebrae, VectorDict
from .barrack import bowyer
from collections import MutableMapping
from warnings import warn



class _Hankyu(InclusiveAdder,dict):
    pass


class Hankyu(InclusiveAdder):
    """Use this at your own risk.
    Hankyu is the same class with the mnemonic for d(efault)dict with addition

 >>> from archery.bow import Hankyu
 >>> tata = Hankyu( dict(a = 1, b = 0, c = -1 ) )
 >>> 
 >>> toto = Hankyu( dict(a = 1 ) )
 >>> print toto+tata
 {'a': 2, 'c': -1, 'b': 0}
 >>> toto-=( tata+tata )
 >>> print toto
 {'a': -1, 'c': 2, 'b': 0}
"""
    mapping = dict
from copy import deepcopy
class Hankyu(_Hankyu, dict):
    """Fix the broken copier
    metaclass are hard"""
    def copy(self):
        return deepcopy(
            bowyer(
                Hankyu,
                self
            )
        )


class _Daikyu(LinearAlgebrae):
    """japanese longbow"""
    """Fix the broken copier
    metaclass are hard"""
    def copy(self):
        return deepcopy(
            bowyer(
                Daikyu,
                self
            )
        )

class Daikyu(_Daikyu, dict):
    """Fix the broken copier
    metaclass are hard"""
    pass


mdict = Daikyu


class _vdict(VectorDict,dict): pass

class vdict(_vdict):
    def __init__(self, *a, **kw):
        super(_vdict, self).__init__(*a, **kw)
        for k, v in self.items():
            if isinstance(v, MutableMapping):
                self[k] = bowyer(globals()[type(self).__name__], v)

class _sdict(Searchable, LinearAlgebrae, dict):
    """japanese longbow"""
    mapping = dict

class sdict(_sdict):
    pass


### for compatibility puprose
class edict(_sdict):

    def __init__(self, *a, **kw):
        warn("ExpDict/edict will become sdict", DeprecationWarning)
        super(_sdict, self).__init__(*a, **kw)
        for k, v in self.items():
            if isinstance(v, MutableMapping):
                self[k] = bowyer(globals()[type(self).__name__], v)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Soon will come sets of consistent adders/subber/muler/diver
grouped in a consistent set known as algebrae
all the combined traits will ensure consistant properties on :
- commutativity;
- associativity;
- distributivity;
- neutral element of addition & multiplication

Planned : 
Linear algebrae,
Hermitian Algebrae 
and so much more....

"""
from .trait import InclusiveAdder, ExclusiveMuler, Vector
from .trait import InclusiveSubber, TaintedExclusiveDiver
from .trait import Copier


class LinearAlgebrae(
    Copier,
    InclusiveAdder, 
    ExclusiveMuler, 
    TaintedExclusiveDiver, 
    InclusiveSubber):
    """A set of + - * / pretty consistant"""
    pass

class SimplyAdd(Copier, InclusiveAdder): pass

class VectorDict(
        LinearAlgebrae,
        Vector
    ):
    pass





#!/usr/bin/env python
# -*- coding: utf8 -*-
"""Making dict great vectors!
"""

from .bow import mdict, vdict, sdict
from .barrack import (
    paired_row_iter, bowyer, make_from_path, Path, mapping_row_iter
)
__version__ = "1.2.0"
__author__ = "julien tayon"
__author_email__ = "julien@tayon.net"
__all__ = [
     "__version__", "vdict","sdict", "mdict", "paired_row_iter", 
     "mapping_row_iter", "__author__", "__author_email__"]






#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Traits / mixins are the heart of archery, but sometimes you need functions to help you

"""
from collections import MutableMapping

MARKER=object()

class Path(tuple):
    def endswith( self, *a_tuple ):
        """check if path ends with the consecutive given has argumenbts value

 >>> p = Path( [ 'a', 'b', 'c' ] )
 >>> p.endswith( 'b', 'c' )
 >>> True
 >>> p.endswith( 'c', 'b' )
 >>> False
        """
        return self[len(self) - len(a_tuple) : ] == a_tuple

    def startswith( self, *a_tuple ):
        """checks if a path starts with the value

 >>> p = Path( [ 'a', 'b', 'c', 'd' ] )
 >>> p.startswith( 'a', 'b' )
 >>> True
        """
        return self[: len( a_tuple ) ] == a_tuple

    def _contains( self, a_tuple, _from = 0, follow = 0):
        if len( a_tuple) == follow:
            return True
        index = False
        here = self[ _from:]

        try:
            index = here.index(a_tuple[follow] )
            return  self._contains(
                    a_tuple, 
                        index +  1 , 
                        follow + 1
                    )
        except ValueError:
            return False
        return False

    def contains(self, *a_tuple ):
        """checks if the serie of keys is contained in a path

 >>> p = Path( [ 'a', 'b', 'c', 'd' ] )
 >>> p.contains( 'b', 'c' )
 >>> True

        """
        return self._contains(a_tuple)

    def value(self):
        """ function provided for code readability:
        - returns the left most value of the Path aka the value
        """
        return self[-1]

    def key(self):
        """ function provided for code readability:
        - returns all the keys in the Path
        """
        return Path(self[:-1])

def make_from_path(type_of_mapping, path):
    """Work in Progress
    create a mutable mapping from a `Path`_ (tuple made of a series of keys in a dict leading to a
    value followed by a value).
    The source is used a mapping factory and is reset in the process

    >>> make_from_path(dict, ("y", "z", 2))
    >>> #Out[2]: {'y': {'z': 2}}

    """
    path = list(path)
    value = path.pop()
    last_key = path.pop()
    tmap = type_of_mapping
    mapping = tmap({last_key : value})
    while path:
        _next = path.pop()
        mapping = tmap({_next : mapping })
    return mapping

def bowyer(_mapping_fact, _mapping_to_convert):
    """the craftsman that makes bow
    A function that given a function in the form 
    f(tree) will convert a one level tree in your mapping"""
    _toc = _mapping_fact(_mapping_to_convert)
    for k, v in _toc.items():
        if isinstance(v, MutableMapping) and len(v):
            _toc[k] = bowyer(_mapping_fact, v)
    return _toc

def mapping_row_iter(tree, path=MARKER):
    """
    iterator on a tree that yield an iterator on a mapping in the form of 
    a list of ordered key that leads to the element and the value

    >>> from archery import mapping_row_iter
    >>> [ x for x in mapping_row_iter({
    ...        "john" : {'math':10.0, 'sport':1.0},~
    ...        "lily" : { 'math':20, 'sport':15.0}
    ...    })]
    >>> #[['john', 'sport', 1.0], ['john', 'math', 10.0],~
    >>> #['lily', 'sport', 15.0], ['lily', 'math', 20]]

    """
    if path is MARKER:
        path = ()
    for k, v in tree.items():
        if isinstance(v, MutableMapping) and len(v):
            for child in mapping_row_iter(v, (path + (k,))):
                yield child
        else:
            yield path + (k , v)


def paired_row_iter(tree, path=MARKER):
    """
    iterator on a tree that yield a pair key (path to a value), value
    """
    if path is MARKER:
        path = tuple()

    for k, v in tree.items():
        if isinstance(v, MutableMapping) and len(v):
            for child in paired_row_iter(v, path + (k,)):
                yield child
        else:
            yield ((path + (k,)) , v)





#/usr/bin/env python
import os
import sys

import warnings
import unittest
from  archery.bow import Hankyu, Daikyu, sdict, vdict, mdict, edict
from archery.barrack import bowyer, Path, make_from_path, paired_row_iter

from math import sqrt, cos, pi, sin, acos

class TestDeprecation(unittest.TestCase):
    def test_deprec_edict(self):
        """ edict (ExpDict is becoming searchable """
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            edict(x=1)
            self.assertTrue(
                issubclass(w[-1].category, DeprecationWarning),
            )

class TestKyleExpectation(unittest.TestCase):

    def test_commutativity(self):
        a = mdict({"a": 1, "y": {"b": 4, "c": [5]}})
        b=  mdict({"a": 1, "x": 5, "y": {"b": 6, "c": [12]}})
        expected_ab = {"a": 2, "x": 5, "y": {"b": 10, "c": [5, 12]}}
        expected_ba = {"a": 2, "x": 5, "y": {"b": 10, "c": [12, 5]}}
        self.assertEqual(a + b, expected_ab)
        a = mdict({"a": 1, "y": mdict({"b": 4, "c": [5]})})
        b=  mdict({"a": 1, "x": 5, "y": mdict({"b": 6, "c": [12]})})
        self.assertEqual(b + a, expected_ba)
        a = mdict({"a": 1, "y": mdict({"b": 4, "c": [5]})})
        b=  mdict({"a": 1, "x": 5, "y": mdict({"b": 6, "c": [12]})})
        a += b
        self.assertEqual(a, expected_ab)


    def test_bug_iadd(self):
        a = mdict({"a": 1, "y": mdict({"b": 4, "c": [5]})})
        b=  mdict({"a": 1, "x": 5, "y": mdict({"b": 6, "c": [12]})})
        expected_ab = {"a": 2, "x": 5, "y": {"b": 10, "c": [5, 12]}}
        expected_ba = {"a": 2, "x": 5, "y": {"b": 10, "c": [12, 5]}}
        self.assertEqual(a + b, expected_ab)
        self.assertEqual(b + a, expected_ba)
        a += b
        self.assertEqual(a, expected_ab)


class TestVectorDict(unittest.TestCase):
    """test a dict with Vector properies (cos, dot, abs)"""

    def setUp(self):
        self.point = vdict( x=3, y=3, z=0)
        self.point2 = vdict( x=1, y=1, z=1)
    
    def testNorm(self):
        self.assertAlmostEqual(
                abs(self.point),
                sqrt(3*3+3*3)
        )
    def testCos(self):
        self.assertAlmostEqual(
                vdict(x=1,y=0).cos(vdict(x=1,y=1)),
                cos(pi/4)
        )

class TestBarrack(unittest.TestCase):
    def test_path(self):
        self.assertEqual(
            Path(("a","b",1)).key(),
            ("a","b")
        )
        self.assertFalse(
            Path(("a","b",1)).contains("a","b","c")
        )
        self.assertEqual(
            Path(("a","b",1)).value(),
            1
        )
        self.assertTrue(Path(("x", "y", 1)).startswith("x"))
        self.assertTrue(Path(("s", "x", "y", 1)).contains("x","y"))

    def test_paired_row_iter(self):
        self.assertEqual(set(paired_row_iter(dict(x=1, a=dict(b=dict(c=1))))),
            { (("x",),1), (("a","b","c"), 1)}
        )


class TestBugWeired(unittest.TestCase):

    def test_stay_same(self):

        class Matrix(mdict):
            def __call__(self, other):
                other = other.copy()
                res= vdict()
                for (src, dst), functor in self.items():
                    res += mdict({ dst: functor(other[src])})
                return res


        theta = pi/6

        u = mdict(x=1, y=2)
        v = mdict(x=1, y=0)
        alien = vdict(x=u, y=v)

        def rotation_maker(theta):
            """"Matrix takes as key (SRC, DST) (which is the opposite of "actual notation")
            """
            return Matrix({
                ("x", "x") : lambda v:1.0 *  v * cos(theta),
                ("y", "x") : lambda v:1.0 * -v * sin(theta),
                ("x", "y") : lambda v:1.0 *  v * sin(theta),
                ("y", "y") : lambda v:1.0 *  v * cos(theta)
            })

        rotation = rotation_maker(pi/6)
        self.assertAlmostEqual(
            rotation(u),
            {'x': -0.13397459621556118, 'y': 2.232050807568877}
            )
# OUT:{'x': 1, 'y': 0}
        self.assertAlmostEqual(rotation(v),
            {'x': 0.8660254037844387, 'y': 0.49999999999999994}
        )
        self.assertAlmostEqual(
            acos(vdict(u).cos(vdict(rotation(u))))/2 / pi * 360,
            29.999999999999993
        )
        self.assertEqual(u, dict(x=1, y=2))
        self.assertAlmostEqual(acos(vdict(v).cos(vdict(rotation_maker(pi/3)(v))))/2 / pi * 360,
60.0)
        self.assertAlmostEqual(
            acos(vdict(v).cos(vdict(rotation_maker(pi/5)(v))))/2 / pi * 360,
            36.0
        )
        self.assertEqual(
            alien,
            {'x': {'x': 1, 'y': 2}, 'y': {'x': 1, 'y': 0}}
        )
        """
        undefined behaviour until I can make a sense of this



        self.assertAlmostEqual(
            rotation_maker(pi/4)(alien),
            {
                'x': {'x': 1.1102230246251565e-16, 'y': 1.4142135623730951},
                'y': {'x': -1.1102230246251565e-16, 'y': 1.414213562373095}
            }
        )
        self.assertAlmostEqual(
            acos(alien.cos(rotation_maker(pi/4)(alien)))/2 / pi * 360,
            54.73561031724534
        )

        self.assertAlmostEqual(rotation_maker(pi/4)(alien),
            {'x':
                {'x': 1.1102230246251565e-16, 'y': 1.4142135623730951},
             'y': {'x': -1.1102230246251565e-16, 'y': 1.414213562373095}})
        """


class TestSearchableDict(unittest.TestCase):
    def setUp(self):
        self.easy = vdict( x=1, y=1, z=0)
        
        self.tree = sdict(
            a = 1,
            b = dict(
                c = 3.0,
                d = dict(e=True)
            ),
            point = self.easy
        )

    def test_rec_type(self):
        """test that if created with another mutale mapping
        does not converts it on the fly to himself"""
        self.assertNotEqual(
            type(self.tree["point"]),
            sdict
        )

    def test_search(self):
        self.assertEqual(
                set([ x for x in self.tree.search(
                    lambda x : Path(x).contains("point","x") 
                        or Path(x).endswith(3.0)
                    )
                ]),
                set([('b', 'c', 3.0), ('point', 'x', 1)])
        )
        self.assertEqual(
                set([ x for x in self.tree.search(
                    lambda x : Path(x).startswith("a")
                    )
                ]),
                set([('a', 1)])
        )

    def test_leaf_search(self):
        self.assertEqual(
                [ i for i in self.tree.leaf_search(lambda x : type(x) is  float)],
                [ 3.0 ]
        )
        self.assertEqual(
                [ i for i in self.tree.leaf_search(lambda x : type(x) is  bool)],
                [ True ]
        )
        


        

class TestHankyu(unittest.TestCase):
    def setUp(self):
        self.easy = Hankyu( x=1, y=1, z=0)
        self.easy_too = Hankyu( x=0, y=-2, o=0)
        
        self.a_tree = dict(
            a = 1,
            b = dict(
                c = 3.0,
                d = dict(e=True)
            ),
            point = self.easy
        )
        self.cplx = bowyer(Hankyu,self.a_tree)

    def test_convert(self):
        self.assertEqual(
            bowyer(Hankyu, {'a': {'b': 1}}),
            Hankyu( {'a': Hankyu(b=1)})
            )
        

    def test_add_not_exists(self):
        self.easy += dict(a=1)
        self.assertEqual(self.easy['a'], 1)
        self.assertEqual(
                self.easy + self.easy_too, 
                sum([ self.easy_too, self.easy])
        )

    def test_add_exists(self):
        self.easy += dict(x=1)
        self.assertEqual(self.easy['x'] , 2)

    def test_number_inc(self):
        self.easy += 1
        self.assertEqual(self.easy['x'] , 2)


    def test_bug_convert_empty_mapping(self):
        """bug #8 : empty mapping intialisation dont work"""
        bug = bowyer( Hankyu,dict( a={}))
        expected = Hankyu( {'a': {}})
        self.assertEqual(
            expected,
            bug
        )

class TestDaikyu(unittest.TestCase):
    def setUp(self):
        self.easy = Daikyu( x=1, y=1, z=0)
        self.easy_too = Daikyu( x=0, y=-2, o=0)
        
        self.a_tree = dict(
            a = 1,
            b = dict(
                c = 3.0,
                d = dict(e=True)
            ),
            point = self.easy
        )
        self.cplx = bowyer(Daikyu,self.a_tree)


    def test_bug_1(self):
        """bug  #1 : * returns a copy where a reference is needed"""
        pt = self.easy
        pt *= -1
        self.assertEqual(
            self.easy,
            pt
        )
    def test_bug_2(self):
        """bug #2 : self * other without affectation should not modify self.
        """
        self.easy *= -1
        a_copy = self.easy.copy()
        self.easy * a_copy
        self.assertEqual(
            self.easy,
            a_copy
        )


    def test_bug_4(self):
        """bug #4 : self / other without affectation should not modify self.
        """
        del(self.easy['z'])
        self.easy /= -1
        a_copy = self.easy.copy()
        self.easy / a_copy
        self.assertEqual(
            self.easy,
            a_copy
        )

    def test_bug_5(self):
        """bug #5 : int * dict without affectation should not modify self.
        """
        a_copy = self.easy.copy()
        2 * self.easy
        self.assertEqual(
            self.easy,
            a_copy
        )

    def test_bug_6(self):
        """bug #6 : int / dict without affectation should not modify self.
        """
        del(self.easy['z'])
        a_copy = self.easy.copy()
        2.0 / self.easy
        self.assertEqual(
            self.easy,
            a_copy
        )


    def test_for_fun(self):
        """just for fun """
        a = self.easy
        b = self.easy_too
        aa = a.copy()
        bb = b.copy()
        self.assertEqual(
            (a - b) * (a + b),
            aa * aa - bb * bb
        )


    def test_for_fun2(self):
        """just for fun """
        a = self.easy
        b = self.easy_too
        aa = a.copy()
        bb = b.copy()
        self.assertEqual(
            -(a + b) * (a + b),
            -1 * ((aa * aa) + 2 * aa * bb + (bb * bb))
        )
    
    def test_dyslexia_bug4(self):
        """counfonding right & left in all r(div/add/sub/mul) operator """
        a = self.easy
        self.assertEqual(
            2/Daikyu({'a':1}),
            {'a' : 2}
        )
        self.assertEqual(
            2-a,
            -(a-2)
        )
    def test_inplace_bug5(self):
        """a/b in place"""
        a = self.easy*4
        self.assertEqual(
            a/{'x':2},
            {'x' : 2}
        )

if __name__ == '__main__':
    unittest.main(verbosity=4)



#/usr/bin/env python
import os
import sys

import unittest
from archery.quiver import VectorDict
from archery.barrack import make_from_path, mapping_row_iter
from archery import mdict
from collections import Counter

from math import sqrt, cos, pi

class CWCos(VectorDict, Counter):
    pass

class TestCosWithCounter(unittest.TestCase):


    def testCos(self):
        self.assertAlmostEqual(
                CWCos(["mot", "wut", "wut", "bla"]).cos(CWCos(["mot","wut", "bla"])),
                0.942809041582
        )
        self.assertAlmostEqual(
                CWCos(["mot", "wut", ]).cos(CWCos(["mot",])),
                cos(pi/4)
        )

class TestBuildingFromMap(unittest.TestCase):

    def test_weired_exp_idea(self):
        tree= dict(a=dict(b=1, c=1, d=dict(e=2, f=dict(e=5)), v=3))
        self.assertEqual(tree,
                sum([ make_from_path(mdict, p) for p in mapping_row_iter(tree)])
        )

if __name__ == '__main__':
    unittest.main(verbosity=4)



#!/usr/bin/env python
# -*- coding: "utf-8" -*-

#from setuptools import setup, find_packages
#import unittest
#import sys
from distutils.command.build_py import build_py as _build_py
from distutils.core import setup
from archery import __version__
from archery import __author__
from archery import __author_email__
import unittest
import sys

def test():
    """Specialized Python source builder."""
    from archery import test_archery
    loader= unittest.TestLoader()
    suite=loader.loadTestsFromModule(test_archery)
    runner=unittest.TextTestRunner()
    result=runner.run(suite)
    if  not result.wasSuccessful():
        raise Exception( "Test Failed: Aborting install")

if "install" in sys.argv or "sdist" in sys.argv or "update" in sys.argv:
    test()

setup(
        name='archery',
        version=__version__,
        author=__author__,
        author_email=__author_email__,
        packages=['archery'],
        url='http://archery.readthedocs.org/',
        license="Python Software Foundation License",
        description='Traits (Mixins) to give +,/,-,* to MutableMapping ',
        long_description=open("README.txt").read(),
        requires=[ ],
        classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: Python Software Foundation License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          ],
)
